/*-
 * Copyright (c) 2006 Stanislav Sedov <stas@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $MBSDlabs: devcpu/cpu/cpu.c,v 1.17 2008/01/25 05:59:10 stas Exp $
 *
 */
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/fcntl.h>
#include <sys/ioccom.h>
#include <sys/malloc.h>
#include <sys/module.h>
#include <sys/mutex.h>
#include <sys/proc.h>
#include <sys/queue.h>
#include <sys/sched.h>
#include <sys/signalvar.h>
#include <sys/kernel.h>
#include <sys/sysctl.h>
#include <sys/uio.h>
#include <sys/pcpu.h>
#include <sys/smp.h>

#include <sys/pmckern.h>

#include <machine/md_var.h>
#include <machine/pcb.h>
#include <machine/specialreg.h>

#include "cpu.h"

static d_open_t cpu_open;
static d_ioctl_t cpu_ioctl;

#define CPU_VERSION 1.1

#ifdef DEBUG
# define DPRINTF(format,...) printf(format, __VA_ARGS__);
#else
# define DPRINTF(...)
#endif

#if __FreeBSD_version < 700052
# define SCHED_LOCK mtx_lock_spin(&sched_lock)
# define SCHED_UNLOCK mtx_unlock_spin(&sched_lock)
#else
# define SCHED_LOCK thread_lock(curthread)
# define SCHED_UNLOCK thread_unlock(curthread)
#endif

int cpu_wrmsr(int, uint64_t);
int cpu_rdmsr(int, uint64_t *);
void msr_onfault(void) __asm(__STRING(msr_onfault));
static int cpu_do_msr(int cpu, cpu_msr_args_t *data, u_long cmd);
static int cpu_do_cpuid(int cpu, cpu_cpuid_args_t *data);
static int cpu_do_update(int cpu, cpu_update_args_t *data);
static int update_intel(int cpu, cpu_update_args_t *args);
static int update_amd(int cpu, cpu_update_args_t *args);

static struct cdev **cpu_devs;
static MALLOC_DEFINE(M_CPU, "cpu", "DEVCPU buffer");

static struct cdevsw cpu_cdevsw = {
        .d_version =    D_VERSION,
        .d_flags =      0,
        .d_open =       cpu_open,
        .d_ioctl =      cpu_ioctl,
        .d_name =       "cpu",
};

/*
 * This function checks if specified cpu enabled or not.
 */
static int
cpu_enabled(int cpu)
{
	return(pmc_cpu_is_disabled(cpu) == 0);
}

/*
 * Check if the current thread is bound to a specific cpu
 */
static int
cpu_sched_is_bound(struct thread *td)
{
	int ret;

	SCHED_LOCK;
	ret = sched_is_bound(td);
	SCHED_UNLOCK;

	return ret;
}

/*
 * Switch to target cpu to run
 */
static void
set_cpu(int cpu)
{
	
	KASSERT(cpu >= 0 && cpu < mp_ncpus && cpu_enabled(cpu), \
		("[cpu,%d]: bad cpu number %d", __LINE__, cpu));

	SCHED_LOCK;
	sched_bind(curthread, cpu);
	SCHED_UNLOCK;
	
	KASSERT(curthread->td_oncpu == cpu, \
		("[cpu,%d]: didn't bound to target cpu %d", __LINE__, cpu));
}

static void
restore_cpu(int cpu, int is_bound)
{

	KASSERT(cpu >= 0 && cpu < mp_ncpus && cpu_enabled(cpu), \
		("[cpu,%d]: bad cpu number %d", __LINE__, cpu));

	SCHED_LOCK;

	if (is_bound == 0)
		sched_unbind(curthread);
	else
		sched_bind(curthread, cpu);
		
	SCHED_UNLOCK;
}

int
cpu_ioctl(struct cdev *dev, u_long cmd, caddr_t data, \
	int flags __unused, struct thread *td)
{
	int ret;
	int cpu = minor(dev);

	if (cpu >= mp_ncpus || !cpu_enabled(cpu)) {
		DPRINTF("[cpu,%d]: bad cpu number %d\n", __LINE__, cpu);
		return ENXIO;
	}

	switch (cmd) {
	case CPU_RDMSR:
	case CPU_WRMSR:
		ret = cpu_do_msr(cpu, (cpu_msr_args_t *)data, cmd);
		break;
	case CPU_CPUID:
		ret = cpu_do_cpuid(cpu, (cpu_cpuid_args_t *)data);
		break;
	case CPU_UPDATE:
		ret = cpu_do_update(cpu, (cpu_update_args_t *)data);
		break;
	default:
		ret = EINVAL;
	}

	return ret;
}

/*
 * Actually perform cpuid operation. No-op on non-x86 compatible hardware.
 */
static int
cpu_do_cpuid(int cpu, cpu_cpuid_args_t *data)
{
	int is_bound = 0;
	int oldcpu;

	KASSERT(cpu >= 0 && cpu < mp_ncpus, \
		("[cpu,%d]: bad cpu number %d", __LINE__, cpu));

	/* Explicitly clear cpuid data to avoid returning stale info */
	bzero(data->data, sizeof(data->data));

	DPRINTF("[cpu,%d]: retriving cpuid level %#0x for %d cpu\n", __LINE__, \
		data->level, cpu);

#if defined(__i386__) || defined(__amd64__)

	oldcpu = curthread->td_oncpu;
	is_bound = cpu_sched_is_bound(curthread);
	set_cpu(cpu);

	critical_enter();
	do_cpuid(data->level, data->data);
	critical_exit();

	restore_cpu(cpu, is_bound);

	return 0;
# endif

	return ENXIO;
}

/*
 * Actually perform MSR operations. No-op on non-x86.
 */
static int
cpu_do_msr(int cpu, cpu_msr_args_t *data, u_long cmd)
{
	int is_bound = 0;
	int oldcpu;
	int ret;

	KASSERT(cpu >= 0 && cpu < mp_ncpus, \
		("[cpu,%d]: bad cpu number %d", __LINE__, cpu));

	/* Explicitly clear cpuid data to avoid returning stale info */
	if (cmd == CPU_RDMSR)
		data->data = 0;

	DPRINTF("[cpu,%d]: operating on MSR %#0x for %d cpu\n", __LINE__, \
		data->msr, cpu);

#if defined(__i386__) || defined(__amd64__)
	oldcpu = curthread->td_oncpu;
	is_bound = cpu_sched_is_bound(curthread);
	set_cpu(cpu);

	critical_enter();
	PCPU_GET(curpcb)->pcb_onfault = (char *)msr_onfault;
	ret = cmd == CPU_RDMSR ? cpu_rdmsr(data->msr, &data->data) : \
				 cpu_wrmsr(data->msr, data->data);
	PCPU_GET(curpcb)->pcb_onfault = (char *)NULL;
	critical_exit();

	restore_cpu(cpu, is_bound);

# else /* i386 || amd64 */
	/* Not yet supported */
	ret = ENXIO;

# endif

	return ret;
}

/*
 * Actually perform microcode update
 */
static int
cpu_do_update(int cpu, cpu_update_args_t *data)
{
	cpu_cpuid_args_t args = {
		.level = 0,
	};
	char vendor[13];
	int ret;

	KASSERT(cpu >= 0 && cpu < mp_ncpus, \
		("[cpu,%d]: bad cpu number %d", __LINE__, cpu));

	DPRINTF("[cpu,%d]: XXX %d", __LINE__, cpu);

#if defined(__i386__) || defined(__amd64__)

	ret = cpu_do_cpuid(cpu, &args);
	if (ret != 0) {
		DPRINTF("[cpu,%d]: can't retrive cpuid info for cpu %d", \
			__LINE__, cpu);
		return ENXIO;
	}
	
	((uint32_t *)vendor)[0] = args.data[1];
	((uint32_t *)vendor)[1] = args.data[3];
	((uint32_t *)vendor)[2] = args.data[2];
	vendor[12] = '\0';

	if (strncmp(vendor, INTEL_VENDOR_ID, strlen(INTEL_VENDOR_ID)) == 0)
		return update_intel(cpu, data);
	else if(strncmp(vendor, INTEL_VENDOR_ID, strlen(AMD_VENDOR_ID)) == 0)
		return update_amd(cpu, data);
	else
		return ENXIO;
		
# endif

	return ENXIO;
}

static int
update_intel(int cpu, cpu_update_args_t *args)
{
	void *ptr = NULL;
	uint64_t rev0, rev1;
	uint32_t tmp[4];
	int is_bound = 0;
	int oldcpu;
	int ret;

	if (args->size == 0 || args->data == NULL) {
		DPRINTF("[cpu,%d]: zero-sized firmware image", __LINE__);
		return EINVAL;
	}

	/* Spec requires 16 byte alignment */
	ptr = contigmalloc(args->size, M_CPU, 0, 0, 0xffffffff, 16, 0);
	if (ptr == NULL) {
		DPRINTF("[cpu,%d]: can't allocate %d bytes of memory", \
			__LINE__, args->size);
		return ENOMEM;
	}

	if (copyin(args->data, ptr, args->size) != 0) {
		DPRINTF("[cpu,%d]: copyin %p->%p of %d bytes failed", \
			__LINE__, args->data, args->ptr, args->size);
		ret = EFAULT;
		goto fail;
	}

	oldcpu = curthread->td_oncpu;
	is_bound = cpu_sched_is_bound(curthread);
	set_cpu(cpu);

	critical_enter();
	cpu_rdmsr(MSR_BIOS_SIGN, &rev0); /* get current micorcode revision */

	/*
	 * Perform update
	 */
	cpu_wrmsr(MSR_BIOS_UPDT_TRIG, (uintptr_t)(ptr));
	cpu_wrmsr(MSR_BIOS_SIGN, 0);

	/* Serialize instruction flow */
	do_cpuid(0, tmp);

	cpu_rdmsr(MSR_BIOS_SIGN, &rev1); /* get new micorcode revision */
	critical_exit();

	restore_cpu(cpu, is_bound);

	if (rev1 > rev0)
		ret = 0;
	else
		ret = EEXIST;

fail:
	if (ptr != NULL)
		contigfree(ptr, args->size, M_CPU);
	
	return ret;
}

static int
update_amd(int cpu, cpu_update_args_t *args)
{
	void *ptr = NULL;
	uint32_t tmp[4];
	int is_bound = 0;
	int oldcpu;
	int ret;

	if (args->size == 0 || args->data == NULL) {
		DPRINTF("[cpu,%d]: zero-sized firmware image", __LINE__);
		return EINVAL;
	}

	/* Spec requires 16 byte alignment */
	ptr = contigmalloc(args->size, M_CPU, 0, 0, 0xffffffff, 16, 0);
	if (ptr == NULL) {
		DPRINTF("[cpu,%d]: can't allocate %d bytes of memory", \
			__LINE__, args->size);
		return ENOMEM;
	}

	if (copyin(args->data, ptr, args->size) != 0) {
		DPRINTF("[cpu,%d]: copyin %p->%p of %d bytes failed", \
			__LINE__, args->data, args->ptr, args->size);
		ret = EFAULT;
		goto fail;
	}

	oldcpu = curthread->td_oncpu;
	is_bound = cpu_sched_is_bound(curthread);
	set_cpu(cpu);

	critical_enter();
	/*
	 * Perform update
	 */
	cpu_wrmsr(MSR_K8_UCODE_UPDATE, (uintptr_t)args->data);

	/* Serialize instruction flow */
	do_cpuid(0, tmp);

	critical_exit();

	restore_cpu(cpu, is_bound);

	ret = 0;

fail:
	if (ptr != NULL)
		contigfree(ptr, args->size, M_CPU);
	
	return ret;
}

int
cpu_open(struct cdev *dev, int flags, int fmt __unused, struct thread *td)
{
	int ret = 0;
	int cpu = minor(dev);

	if (cpu >= mp_ncpus || !cpu_enabled(cpu)) {
		DPRINTF("[cpu,%d]: incorrect cpu number %d\n", __LINE__, cpu);
		return ENXIO;
	}

	if (flags & FWRITE)
		ret = securelevel_gt(td->td_ucred, 0);

	return (ret);
}

static int
cpu_modevent(module_t mod __unused, int type, void *data __unused)
{
	int cpu;

	switch(type) {
	case MOD_LOAD:

		/* XXX: should move out checks */
		if ((cpu_feature & CPUID_MSR) == 0) {
			if (bootverbose)
				printf("devcpu: not available\n");
			return ENOSYS;
		}

		if (bootverbose)
			printf("cpu: access to MSR registers/cpuid info\n");

		cpu_devs = (struct cdev **)malloc(sizeof(void *) * mp_ncpus,\
						M_CPU, M_WAITOK | M_ZERO);
		if (cpu_devs == NULL) {
			DPRINTF("[cpu,%d]: cannot allocate memory\n", __LINE__);
			return ENOSYS;
		}

		for (cpu = 0; cpu < mp_ncpus; cpu++)
			if (cpu_enabled(cpu))
				cpu_devs[cpu] = make_dev(&cpu_cdevsw, cpu,
					UID_ROOT, GID_KMEM, 0640,
					"cpu%d", cpu);
		break;

	case MOD_UNLOAD:
		for (cpu = 0; cpu < mp_ncpus; cpu++) {
			if (cpu_devs[cpu] != NULL)
				destroy_dev(cpu_devs[cpu]);
		}

		free(cpu_devs, M_CPU);

		break;

	case MOD_SHUTDOWN:
		break;

	default:
		return(EOPNOTSUPP);
        }

	return (0);
}

DEV_MODULE(cpu, cpu_modevent, NULL);
MODULE_VERSION(cpu, CPU_VERSION);
