/*-
 * Copyright (c) 2006 Stanislav Sedov <stas@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $MBSDlabs: devcpu/cpu_microcode_tool/amd.c,v 1.4 2006/12/08 22:34:03 stas Exp $
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <err.h>
#include <errno.h>
#include <assert.h>
#include <sysexits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include "sys/queue.h"

#include <sys/ioctl.h>
#include <sys/ioccom.h>
#include <cpu/cpu.h>

#include <machine/cpufunc.h>
#include <machine/specialreg.h>

#include "cpu_microcode_tool.h"
#include "amd.h"

struct cpu {
	int			unit;		/* number of cpu */
	uint32_t		signature;
	SLIST_ENTRY(cpu)	next;
};

static SLIST_HEAD(, cpu)	cpu_list = SLIST_HEAD_INITIALIZER(&cpu_list);

static int do_update	__P((int unit, void *image, size_t size));

int
amd_add(unit)
	int unit;
{ 
	struct cpu *cpu = NULL; /* to eleminate gcc warning */
	char buffer[16];
	char vendor[13];
	int fd;
	int ret, retval = 0;

	cpu_cpuid_args_t idargs = {
		.level  = 0,
	};

	snprintf(buffer, 16, "/dev/cpu%d", unit);
	fd = open(buffer, O_RDWR, 0);
	if (fd < 0) {
		WARN(1, "open(%s)", buffer);
		retval = 1;
		goto fail;
	}

	ret = ioctl(fd, CPU_CPUID, &idargs);
	if (ret < 0) {
		WARN(1, "ioctl(%s, %ld)", buffer, CPU_CPUID);
		retval = 1;
		goto fail;
	}

	((uint32_t *)vendor)[0] = idargs.data[1];
	((uint32_t *)vendor)[1] = idargs.data[3];
	((uint32_t *)vendor)[2] = idargs.data[2];
	vendor[12] = '\0';

	if (strncmp(vendor, AMD_VENDOR_ID, strlen(AMD_VENDOR_ID)) != 0)
		goto fail;

	cpu = (struct cpu *)malloc(sizeof(struct cpu));
	if (cpu == NULL) {
		WARN(1, "malloc()");
		goto fail;
	}

	idargs.level = 1; /* Request signature */

	ret = ioctl(fd, CPU_CPUID, &idargs);
	if (ret < 0) {
		WARN(1, "ioctl(%s, %ld)", buffer, CPU_CPUID);
		goto fail;
	}
	
	cpu->signature = idargs.data[0];

	cpu->unit = unit;

	SLIST_INSERT_HEAD(&cpu_list, cpu, next);
	
	WARNX(1, "scheduling cpu %d family %#x model %#x "		\
		"stepping %#x extfamily %#x extmodel %#x for update",	\
		unit, (cpu->signature >> 8) & 0x0f,			\
		(cpu->signature >> 4) & 0x0f,				\
		(cpu->signature >> 0) & 0x0f,				\
		(cpu->signature >> 20) & 0xff,				\
		(cpu->signature >> 16) & 0x0f);

fail:
	if (cpu != NULL && retval != 0)
		free(cpu);

	if (fd >= 0)
		close(fd);

	return retval;
}

void
amd_free()
{

	SLIST_FREE(&cpu_list, next, free);
}

void
amd_update(path)
	const char *path;
{
	int fd;
	struct stat st;
	uint32_t *fw_image = MAP_FAILED;
	uint32_t sum;
	unsigned int i;
	amd_fw_header_t *fw_header;
	int ret;
	uint32_t *fw_data;
	size_t fw_size;
	struct cpu *cpu;

	ASSERT(path);

	/*
	 * Open firmware file
	 */
	fd = open(path, O_RDONLY, 0);
	if (fd < 0) {
		WARN(1, "open(%s)", path);
		goto fail;
	}

	ret = fstat(fd, &st);
	if (ret != 0) {
		WARN(1, "fstat()");
		goto fail;
	}

	if (st.st_size < 0 || (unsigned)st.st_size < sizeof(amd_fw_header_t)) {
		WARNX(2, "file too short: %s", path);
		goto fail;
	}

	/*
	 * MMAP the whole image
	 */
	fw_image = (uint32_t *)mmap(NULL, st.st_size, PROT_READ, \
						MAP_PRIVATE, fd, 0);
	if  (fw_image == MAP_FAILED) {
		WARN(1, "mmap(%s)", path);
		goto fail;
	}

	fw_header = (amd_fw_header_t *)fw_image;
	if ((fw_header->magic >> 8) != AMD_MAGIC) {
		WARNX(2, "%s isn't valid amd firmware: version mismatch", \
			path);
		WARNX(2, "%#x", fw_header->magic);
		goto fail;
	}

	fw_data = (uint32_t *)&(fw_header[1]);
	fw_size = (st.st_size - sizeof(amd_fw_header_t))/sizeof(uint32_t);

	/*
	 * Check primary checksum
	 */
	sum = 0;
	for (i = 0; i < fw_size; i++)
		sum += fw_data[i];
	if (sum != fw_header->checksum) {
		WARNX(2, "%s: update data checksum invalid", path);
		goto fail;
	}

	/*
	 * Find matched cpus
	 */
	SLIST_FOREACH(cpu, &cpu_list, next) {
		ASSERT(cpu);

		if (cpu->signature == fw_header->signature) {
			WARNX(1, "%s: updating cpu %d", path, cpu->unit);

			ret = do_update(cpu->unit, fw_image, st.st_size);
			if (ret == 0) {
				WARNX(1, "%s: done", path);
			}
			else {
				WARNX(1, "%s: failed", path);
			}
		}
	}

fail:
	if (fw_image != MAP_FAILED)
		if(munmap(fw_image, st.st_size) != 0)
			warn("munmap(%s)", path);

	if (fd >= 0)
		close(fd);
}

/*
 * Actually perform update
 */
static int
do_update(unit, image, size)
	int	unit;
	void	*image;
	size_t	size;
{
	char buffer[16];
	int fd;
	int ret;

	ASSERT(image);

	cpu_update_args_t args;

	snprintf(buffer, 16, "/dev/cpu%d", unit);
	fd = open(buffer, O_RDWR, 0);
	if (fd < 0) {
		WARN(0,"open(%s)", buffer);
		ret = 1;
		goto fail;
	}

	args.data = image;
	args.size = size;
	ret = ioctl(fd, CPU_UPDATE, &args);
	if (ret < 0) {
		WARN(0, "ioctl(%s, %ld)", buffer, CPU_UPDATE);
		goto fail;
	}

fail:
	if (fd >= 0)
		close(fd);

	return ret;
}
