/*-
 * Copyright (c) 2006 Stanislav Sedov <stas@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $MBSDlabs: devcpu/cpu_microcode_tool/cpu_microcode_tool.c,v 1.7 2006/12/08 23:23:28 stas Exp $
 *
 */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <err.h>
#include <sysexits.h>
#include <assert.h>
#include <dirent.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/endian.h>
#include <sys/stat.h>
#include <sys/queue.h>

#include <sys/sysctl.h>

#include "cpu_microcode_tool.h"
#include "intel.h"
#include "amd.h"

/* Global data */
int verbose_level = 0;

/* Macros for freeing SLISTs, probably must be in /sys/queue.h */
#define SLIST_FREE(head, field, freef) do {		\
    typeof(SLIST_FIRST(head)) __elm0;			\
    typeof(SLIST_FIRST(head)) __elm;			\
    SLIST_FOREACH_SAFE(__elm, (head), field, __elm0)	\
        (void)(freef)(__elm);				\
} while(0);

struct {
	mu_update_t	*update;
	mu_add_t	*add_cpu;
	mu_free_t	*free;
} fw_handlers[] = {
	{intel_update, intel_add, intel_free},
	{amd_update, amd_add, amd_free},
};

#define NHANDLERS (sizeof(fw_handlers) / sizeof(*fw_handlers))

struct fw {
	char		*path;
	SLIST_ENTRY(fw)	next;
};

#define AFLAG 0x0001

static SLIST_HEAD(, fw) fw_list = SLIST_HEAD_INITIALIZER(&fw_list);
static void		usage			__P((void));
int			main			__P((int argc, char **argv));
static int		add_datadir		__P((char *path));
static int		isdir			__P((char *path));

static void
usage(void)
{

	fprintf(stderr, "usage: %s [-I dir] [-avh] cpu\n", getprogname());

	exit(EX_USAGE);
}

static void
fw_free(struct fw *fw)
{
	
	ASSERT(fw);
	ASSERT(fw->path);
	free(fw->path);
	free(fw);
}

static int
isdir(path)
	char	*path;
{
	int ret;
	struct stat st;

	ret = stat(path, &st);
	if (ret < 0) {
		WARN(0, "stat(%s)", path);
		return ret;
	}

	return (st.st_mode & S_IFDIR);
}

static int
add_datadir(path)
	char	*path;
{
	DIR *dir;
	struct fw *fw;
	struct dirent *dirent;
	int ret;

	dir = opendir(path);
	if (dir == NULL) {
		WARNX(1, "skipping directory %s: not accessible\n", path);
		return 1;
	}

	while ((dirent = readdir(dir)) != NULL) {

		if (dirent->d_namlen == 0)
			continue;

		fw = (struct fw *)malloc(sizeof(struct fw));
		if (fw == NULL) {
			WARNX(0, "can't allocate memory");
			continue;
		}
		ret = asprintf(&fw->path, "%s/%s", path, dirent->d_name);
		if (ret < 0) {
			WARNX(0, "can't allocate memory");
			continue;
		}

		if (isdir(fw->path) != 0) {
			free(fw->path);
			free(fw);
			continue;
		}

		SLIST_INSERT_HEAD(&fw_list, fw, next);
	}
	(void)closedir(dir);

	return 0;
}

int
main(argc, argv)
	int argc;
	char *argv[];
{
	char ch;
	int flags = 0;
	unsigned int unit;
	unsigned int i, j;
	struct fw *fw;
	int ret;

	while ((ch = getopt(argc, argv, "avhI:")) != -1) {
		switch (ch) {
		case 'a':
			flags |= AFLAG;
			break;

		case 'I':
			(void)add_datadir(optarg);
			break;

		case 'v':
			verbose_level++;
			break;

		case 'h':
			/* FALLTHROUGH */
		
		default:
			usage();
			/* NOTREACHED */
		}
	}
	argc -= optind;
	argv += optind;

	if (argc < 1 && (flags & AFLAG) == 0)
		usage();

	if (flags & AFLAG) { /* If we want to update all cpus */
		unsigned int ncpu;
		size_t buflen = sizeof(ncpu);
		ret = sysctlbyname("hw.ncpu", &ncpu, &buflen, NULL, 0);
		if (ret != 0) {
			warn("sysctl(\"hw.ncpu\")");
			exit(EX_OSERR);
		}

		for (i = 0; i < ncpu; i++)
			for (j = 0; j < NHANDLERS; j++)
				if (fw_handlers[j].add_cpu(i) != 0)
					break;
	}
	else {
		for (; argc > 0; argc--, argv++) {
			if (!(sscanf(argv[0], "cpu%d", &unit) == 1 || \
				sscanf(argv[0], "/dev/cpu%d", &unit) == 1))
				errx(EX_USAGE, "can't find device %s", argv[0]);
			
				for (i = 0; i < NHANDLERS; i++)
					if (fw_handlers[i].add_cpu(unit) != 0)
						break;
		}
	}

	/*
	 * Try to update with every firmware image
	 */
	for (i = 0; i < NHANDLERS; i++) {
		SLIST_FOREACH(fw, &fw_list, next) {
			(void)fw_handlers[i].update(fw->path);
		}
		fw_handlers[i].free();
	}

	SLIST_FREE(&fw_list, next, fw_free);

	return EX_OK;
}
