/*-
 * Copyright (c) 2006 Stanislav Sedov <stas@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $MBSDlabs: devcpu/cpu_microcode_tool/intel.c,v 1.13 2006/12/08 22:34:03 stas Exp $
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <err.h>
#include <errno.h>
#include <assert.h>
#include <sysexits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include "sys/queue.h"

#include <sys/ioctl.h>
#include <sys/ioccom.h>
#include <cpu/cpu.h>

#include <machine/cpufunc.h>
#include <machine/specialreg.h>

#include "cpu_microcode_tool.h"
#include "intel.h"

#define DEFAULT_UCODE_SIZE 2000 /* size of update data if not specified */

struct cpu {
	int			unit;		/* number of cpu */
	int32_t			revision;	/* firmware revision */
	uint32_t		signature;
	uint32_t		flag;
	SLIST_ENTRY(cpu)	next;
};

static SLIST_HEAD(, cpu)	cpu_list = SLIST_HEAD_INITIALIZER(&cpu_list);

static int do_update	__P((int unit, void *image, size_t size));

int
intel_add(unit)
	int unit;
{ 
	struct cpu *cpu = NULL; /* to eleminate gcc warning */
	char buffer[16];
	char vendor[13];
	int fd;
	int ret, retval = 0;

	cpu_cpuid_args_t idargs = {
		.level  = 0,
	};
	cpu_msr_args_t msrargs = {
		.msr = MSR_IA32_PLATFORM_ID,
	};

	snprintf(buffer, 16, "/dev/cpu%d", unit);
	fd = open(buffer, O_RDWR, 0);
	if (fd < 0) {
		WARN(1, "open(%s)", buffer);
		retval = 1;
		goto fail;
	}

	ret = ioctl(fd, CPU_CPUID, &idargs);
	if (ret < 0) {
		WARN(1, "ioctl(%s, %ld)", buffer, CPU_CPUID);
		retval = 1;
		goto fail;
	}

	((uint32_t *)vendor)[0] = idargs.data[1];
	((uint32_t *)vendor)[1] = idargs.data[3];
	((uint32_t *)vendor)[2] = idargs.data[2];
	vendor[12] = '\0';

	if (strncmp(vendor, INTEL_VENDOR_ID, strlen(INTEL_VENDOR_ID)) != 0)
		goto fail;

	cpu = (struct cpu *)malloc(sizeof(struct cpu));
	if (cpu == NULL) {
		WARN(1, "malloc()");
		goto fail;
	}

	idargs.level = 1; /* Request signature */

	ret = ioctl(fd, CPU_CPUID, &idargs);
	if (ret < 0) {
		WARN(1, "ioctl(%s, %ld)", buffer, CPU_CPUID);
		goto fail;
	}
	
	cpu->signature = idargs.data[0];

	ret = ioctl(fd, CPU_RDMSR, &msrargs);
	if (ret < 0) {
		WARN(1, "ioctl(%s, %ld)", buffer, CPU_RDMSR);
		goto fail;
	}

	/* MSR_IA32_PLATFORM_ID contains flag in BCD in bits 52-50 */
	cpu->flag = 1 << ((msrargs.data >> 50) & 7);

	msrargs.msr = MSR_BIOS_SIGN;
	ret = ioctl(fd, CPU_RDMSR, &msrargs);
	if (ret < 0) {
		WARN(1, "ioctl(%s, %ld)", buffer, CPU_RDMSR);
		goto fail;
	}

	cpu->revision = msrargs.data >> 32; /* revision in the high dword */
	cpu->unit = unit;

	SLIST_INSERT_HEAD(&cpu_list, cpu, next);
	
	WARNX(1, "scheduling cpu %d type %#x family %#x model %#x "	\
		"stepping %#x for update", unit,			\
		(cpu->signature >> 12) & 0x03,				\
		(cpu->signature >> 8) & 0x0f,				\
		(cpu->signature >> 4) & 0x0f,				\
		(cpu->signature >> 0) & 0x0f);

fail:
	if (cpu != NULL && retval != 0)
		free(cpu);

	if (fd >= 0)
		close(fd);

	return retval;
}

void
intel_free()
{

	SLIST_FREE(&cpu_list, next, free);
}

void
intel_update(path)
	const char *path;
{
	int fd;
	struct stat st;
	uint32_t *fw_image = MAP_FAILED;
	intel_ext_header_t *ext_header = NULL;
	ssize_t ext_size;
	int have_ext_table;
	uint32_t sum;
	unsigned int i;
	size_t payload_size;
	intel_fw_header_t *fw_header;
	intel_cpu_signature_t *ext_table = NULL;
	int ret;
	size_t ext_table_size;
	void *fw_data;
	struct cpu *cpu;
	size_t data_size, total_size;

	ASSERT(path);

	/*
	 * Open firmware file
	 */
	fd = open(path, O_RDONLY, 0);
	if (fd < 0) {
		WARN(1, "open(%s)", path);
		goto fail;
	}

	ret = fstat(fd, &st);
	if (ret != 0) {
		WARN(1, "fstat()");
		goto fail;
	}

	if (st.st_size < 0 ||(unsigned)st.st_size < sizeof(intel_fw_header_t)) {
		WARNX(2, "file too short: %s", path);
		goto fail;
	}

	/*
	 * MMAP the whole image
	 */
	fw_image = (uint32_t *)mmap(NULL, st.st_size, PROT_READ, \
						MAP_PRIVATE, fd, 0);
	if  (fw_image == MAP_FAILED) {
		WARN(1, "mmap(%s)", path);
		goto fail;
	}

	fw_header = (intel_fw_header_t *)fw_image;
	if (fw_header->header_version != INTEL_HEADER_VERSION || \
	    fw_header->loader_revision != INTEL_LOADER_REVISION) {
		WARNX(2, "%s isn't valid intel firmware: version mismatch", \
			path);
		goto fail;
	}

	/*
	 * According to spec, if data_size == 0, size of ucode = 2000
	 */
	if (fw_header->data_size == 0)
		data_size = DEFAULT_UCODE_SIZE;
	else
		data_size = fw_header->data_size;

	if (fw_header->total_size == 0)
		total_size = data_size + sizeof(intel_fw_header_t);
	else
		total_size = fw_header->total_size;

	if (total_size > (unsigned)st.st_size || st.st_size < 0) {
		WARNX(1, "file too short: %s", path);
		goto fail;
	}

	payload_size = data_size + sizeof(intel_fw_header_t);

	/*
	 * Check primary checksum
	 */
	sum = 0;
	for (i = 0; i < (payload_size / sizeof(uint32_t)); i++)
		sum += ((uint32_t *)fw_image)[i];
	if (sum != 0) {
		WARNX(2, "%s: update data checksum invalid", path);
		goto fail;
	}

	/*
	 * Check if there is an extended signature table
	 */
	ext_size = total_size - payload_size;

	have_ext_table = 0;

	if (ext_size > (signed)sizeof(intel_ext_header_t)) {
	
		ext_header = \
			(intel_ext_header_t *)&(((char *)fw_image)[payload_size]);
		ext_table = \
			(intel_cpu_signature_t *)&(ext_header[1]);

		/*
		 * Check ext table size
		 */
		ext_table_size = sizeof(intel_ext_header_t) + \
			ext_header->sig_count * sizeof(intel_cpu_signature_t);
		if (ext_table_size + payload_size > total_size) {
			WARNX(1, "%s: broken extended signature table", path);
			goto no_table;
		}

		/*
		 * Check extended table signature
		 */
		sum = 0;
		for (i = 0; i < (ext_table_size / sizeof(uint32_t)); i++)
			sum += ((uint32_t *)ext_header)[i];

		if (sum != 0) {
			WARNX(2, "%s: ext signature table checksum invalid", \
									path);
			goto no_table;
		}

		have_ext_table = 1;
	}

no_table:

	fw_data = &(fw_header[1]); /* pointer to the update data */

	/*
	 * Find matched cpus
	 */
	SLIST_FOREACH(cpu, &cpu_list, next) {
		ASSERT(cpu);

		if (cpu->signature == fw_header->cpu_signature && \
		    (cpu->flag & fw_header->cpu_flags) != 0)
			goto matched;

		if (have_ext_table != 0) {
			ASSERT(ext_header);
			ASSERT(ext_table);
			for (i = 0; i < ext_header->sig_count; i++) {
				uint32_t sig = ext_table[i].cpu_signature;

				if (cpu->signature == sig && \
				    (cpu->flag & ext_table[i].cpu_flags) != 0)
					goto matched;
			}
		}
		continue;

	matched:
		if (cpu->revision >= fw_header->revision) {
			WARNX(1, "%s rev %#x: skipping cpu%d rev %#x " \
				"(up to date)", path, fw_header->revision, \
				cpu->unit, cpu->revision);
			continue;
		}

		WARNX(1, "%s rev %#x: updating cpu %d from rev %#x...", \
			path, fw_header->revision, cpu->unit, cpu->revision);

		ret = do_update(cpu->unit, fw_data, data_size);
		if (ret == 0) {
			WARNX(1, "%s rev %#x: done", path, \
				fw_header->revision);
		}
		else {
			WARNX(1, "%s rev %#x: failed", path, \
				fw_header->revision);
		}
	}

fail:
	if (fw_image != MAP_FAILED)
		if(munmap(fw_image, st.st_size) != 0)
			warn("munmap(%s)", path);

	if (fd >= 0)
		close(fd);
}

/*
 * Actually perform update
 */
static int
do_update(unit, image, size)
	int	unit;
	void	*image;
	size_t	size;
{
	char buffer[16];
	int fd;
	int ret;

	ASSERT(image);

	cpu_update_args_t args;

	snprintf(buffer, 16, "/dev/cpu%d", unit);
	fd = open(buffer, O_RDWR, 0);
	if (fd < 0) {
		WARN(0,"open(%s)", buffer);
		ret = 1;
		goto fail;
	}

	args.data = image;
	args.size = size;
	ret = ioctl(fd, CPU_UPDATE, &args);
	if (ret < 0) {
		WARN(0, "ioctl(%s, %ld)", buffer, CPU_UPDATE);
		goto fail;
	}

fail:
	if (fd >= 0)
		close(fd);

	return ret;
}
